// Generated by CoffeeScript 2.7.0
(function() {
  // A match list is an array of indexes to characters that match.
  // This file should closely follow `scorer` except that it returns an array
  // of indexes instead of a score.
  var basenameMatch, computeMatch, isMatch, isWordStart, match, mergeMatches, scoreAcronyms, scoreCharacter, scoreConsecutives;

  ({isMatch, isWordStart, scoreConsecutives, scoreCharacter, scoreAcronyms} = require('./scorer'));

  
  // Main export

  // Return position of character which matches
  exports.match = match = function(string, query, options) {
    var allowErrors, baseMatches, matches, pathSeparator, preparedQuery, string_lw;
    ({allowErrors, preparedQuery, pathSeparator} = options);
    if (!(allowErrors || isMatch(string, preparedQuery.core_lw, preparedQuery.core_up))) {
      return [];
    }
    string_lw = string.toLowerCase();
    // Full path results
    matches = computeMatch(string, string_lw, preparedQuery);
    if (matches.length === 0) {
      //if there is no matches on the full path, there should not be any on the base path either.
      return matches;
    }
    // Is there a base path ?
    if (string.indexOf(pathSeparator) > -1) {
      // Base path results
      baseMatches = basenameMatch(string, string_lw, preparedQuery, pathSeparator);
      // Combine the results, removing duplicate indexes
      matches = mergeMatches(matches, baseMatches);
    }
    return matches;
  };

  
  // Wrap

  // Helper around match if you want a string with result wrapped by some delimiter text
  exports.wrap = function(string, query, options) {
    var matchIndex, matchPos, matchPositions, output, strPos, tagClass, tagClose, tagOpen;
    if ((options.wrap != null)) {
      ({tagClass, tagOpen, tagClose} = options.wrap);
    }
    if (tagClass == null) {
      tagClass = 'highlight';
    }
    if (tagOpen == null) {
      tagOpen = '<strong class="' + tagClass + '">';
    }
    if (tagClose == null) {
      tagClose = '</strong>';
    }
    if (string === query) {
      return tagOpen + string + tagClose;
    }
    //Run get position where a match is found
    matchPositions = match(string, query, options);
    //If no match return as is
    if (matchPositions.length === 0) {
      return string;
    }
    //Loop over match positions
    output = '';
    matchIndex = -1;
    strPos = 0;
    while (++matchIndex < matchPositions.length) {
      matchPos = matchPositions[matchIndex];
      // Get text before the current match position
      if (matchPos > strPos) {
        output += string.substring(strPos, matchPos);
        strPos = matchPos;
      }
      // Get consecutive matches to wrap under a single tag
      while (++matchIndex < matchPositions.length) {
        if (matchPositions[matchIndex] === matchPos + 1) {
          matchPos++;
        } else {
          matchIndex--;
          break;
        }
      }
      //Get text inside the match, including current character
      matchPos++;
      if (matchPos > strPos) {
        output += tagOpen;
        output += string.substring(strPos, matchPos);
        output += tagClose;
        strPos = matchPos;
      }
    }
    //Get string after last match
    if (strPos <= string.length - 1) {
      output += string.substring(strPos);
    }
    //return wrapped text
    return output;
  };

  basenameMatch = function(subject, subject_lw, preparedQuery, pathSeparator) {
    var basePos, depth, end;
    // Skip trailing slashes
    end = subject.length - 1;
    while (subject[end] === pathSeparator) {
      end--;
    }
    // Get position of basePath of subject.
    basePos = subject.lastIndexOf(pathSeparator, end);
    //If no PathSeparator, no base path exist.
    if (basePos === -1) {
      return [];
    }
    // Get the number of folder in query
    depth = preparedQuery.depth;
    // Get that many folder from subject
    while (depth-- > 0) {
      basePos = subject.lastIndexOf(pathSeparator, basePos - 1);
      if (basePos === -1) { //consumed whole subject ?
        return [];
      }
    }
    
    // Get basePath match
    basePos++;
    end++;
    return computeMatch(subject.slice(basePos, end), subject_lw.slice(basePos, end), preparedQuery, basePos);
  };

  
  // Combine two matches result and remove duplicate
  // (Assume sequences are sorted, matches are sorted by construction.)

  mergeMatches = function(a, b) {
    var ai, bj, i, j, m, n, out;
    m = a.length;
    n = b.length;
    if (n === 0) {
      return a.slice();
    }
    if (m === 0) {
      return b.slice();
    }
    i = -1;
    j = 0;
    bj = b[j];
    out = [];
    while (++i < m) {
      ai = a[i];
      while (bj <= ai && ++j < n) {
        if (bj < ai) {
          out.push(bj);
        }
        bj = b[j];
      }
      out.push(ai);
    }
    while (j < n) {
      out.push(b[j++]);
    }
    return out;
  };

  //----------------------------------------------------------------------

  // Align sequence (used for fuzzaldrin.match)
  // Return position of subject characters that match query.

  // Follow closely scorer.computeScore.
  // Except at each step we record what triggered the best score.
  // Then we trace back to output matched characters.

  // Differences are:
  // - we record the best move at each position in a matrix, and finish by a traceback.
  // - we reset consecutive sequence if we do not take the match.
  // - no hit miss limit
  computeMatch = function(subject, subject_lw, preparedQuery, offset = 0) {
    var DIAGONAL, LEFT, STOP, UP, acro_score, align, backtrack, csc_diag, csc_row, csc_score, i, j, m, matches, move, n, pos, query, query_lw, score, score_diag, score_row, score_up, si_lw, start, trace;
    query = preparedQuery.query;
    query_lw = preparedQuery.query_lw;
    m = subject.length;
    n = query.length;
    //this is like the consecutive bonus, but for camelCase / snake_case initials
    acro_score = scoreAcronyms(subject, subject_lw, query, query_lw).score;
    //Init
    score_row = new Array(n);
    csc_row = new Array(n);
    // Directions constants
    STOP = 0;
    UP = 1;
    LEFT = 2;
    DIAGONAL = 3;
    //Traceback matrix
    trace = new Array(m * n);
    pos = -1;
    //Fill with 0
    j = -1; //0..n-1
    while (++j < n) {
      score_row[j] = 0;
      csc_row[j] = 0;
    }
    i = -1; //0..m-1
    while (++i < m) { //foreach char si of subject
      score = 0;
      score_up = 0;
      csc_diag = 0;
      si_lw = subject_lw[i];
      j = -1; //0..n-1
      while (++j < n) { //foreach char qj of query
        
        //reset score
        csc_score = 0;
        align = 0;
        score_diag = score_up;
        //Compute a tentative match
        if (query_lw[j] === si_lw) {
          start = isWordStart(i, subject, subject_lw);
          // Forward search for a sequence of consecutive char
          csc_score = csc_diag > 0 ? csc_diag : scoreConsecutives(subject, subject_lw, query, query_lw, i, j, start);
          // Determine bonus for matching A[i] with B[j]
          align = score_diag + scoreCharacter(i, j, start, acro_score, csc_score);
        }
        //Prepare next sequence & match score.
        score_up = score_row[j];
        csc_diag = csc_row[j];
        //In case of equality, moving UP get us closer to the start of the candidate string.
        if (score > score_up) {
          move = LEFT;
        } else {
          score = score_up;
          move = UP;
        }
        // Only take alignment if it's the absolute best option.
        if (align > score) {
          score = align;
          move = DIAGONAL;
        } else {
          //If we do not take this character, break consecutive sequence.
          // (when consecutive is 0, it'll be recomputed)
          csc_score = 0;
        }
        score_row[j] = score;
        csc_row[j] = csc_score;
        trace[++pos] = score > 0 ? move : STOP;
      }
    }
    // -------------------
    // Go back in the trace matrix
    // and collect matches (diagonals)
    i = m - 1;
    j = n - 1;
    pos = i * n + j;
    backtrack = true;
    matches = [];
    while (backtrack && i >= 0 && j >= 0) {
      switch (trace[pos]) {
        case UP:
          i--;
          pos -= n;
          break;
        case LEFT:
          j--;
          pos--;
          break;
        case DIAGONAL:
          matches.push(i + offset);
          j--;
          i--;
          pos -= n + 1;
          break;
        default:
          backtrack = false;
      }
    }
    matches.reverse();
    return matches;
  };

}).call(this);
